local EmptyFn = function() end

local Value = class()
function Value:init(data, children, label)
    self.data = data
    self.grad = 0
    self._backward = EmptyFn
    self.prev = children or {}
    self.label = label
end

function Value:__tostring()
    if self.label then
        return string.format("Value(data=%.4f, grad=%.4f, label=%s)", self.data, self.grad, self.label)
    else
        return string.format("Value(data=%.4f, grad=%.4f)", self.data, self.grad)
    end
end

function Value:__add(other)
    if getmetatable(self) ~= Value then
        self = Value(self)
    end
    if getmetatable(other) ~= Value then
        other = Value(other)
    end
    
    local out = Value(self.data + other.data, { self, other })
    
    out._backward = function()
        self.grad = self.grad + out.grad
        other.grad = other.grad + out.grad
    end
    
    return out
end

function Value:__mul(other)
    if getmetatable(self) ~= Value then
        self = Value(self)
    end
    if getmetatable(other) ~= Value then
        other = Value(other)
    end
    
    local out = Value(self.data * other.data, { self, other })
    
    out._backward = function()
        self.grad = self.grad + (other.data * out.grad)
        other.grad = other.grad + (self.data * out.grad)
    end
    
    return out
end

function Value:__pow(other)
    if getmetatable(self) ~= Value then
        self = Value(self)
    end
    
    local out = Value(self.data ^ other, { self })
    
    out._backward = function()
        self.grad = self.grad + ((other * self.data^(other-1)) * out.grad)
    end
    
    return out
end

function Value:__unm()
    return self * -1
end

function Value:__sub(other)
    return self + (-other)
end

function Value:tanh()
    local x = self.data
    local t = (math.exp(2*x) - 1) / (math.exp(2*x) + 1)
    local out = Value(t, { self })
    
    out._backward = function()
        self.grad = self.grad + ((1 - t^2) * out.grad)
    end
    
    return out
end

function Value:backward()
    
    -- Topological order of all children in the graph
    local topo = {}
    local visited = {}
    local build_topo -- decl.
    build_topo = function(v) -- def.
        if visited[v] == nil then
            visited[v] = true
            for _,child in ipairs(v.prev) do
                build_topo(child)
            end
            table.insert(topo, v)
        end
    end
    build_topo(self)
    
    -- Go one variable at a time and apply chain rule to get its gradient
    self.grad = 1
    for i = #topo,1,-1 do
        topo[i]._backward()
    end
end

AI.Value = Value