local Value = AI.Value

local Neuron = class()
function Neuron:init(nin)
    self.w = arraycomp(function()
        -- High precision 64 bit float
        return Value(math.random(0) / math.maxinteger)
    end, range(nin))
    self.b = Value(0)
end

function Neuron:__call(x)
    local act = self.b
    for wi,xi in zip(self.w, x) do
        act = act + wi*xi
    end
    return act:tanh()
end

function Neuron:parameters()
    local out = {}
    table.move(self.w, 1, #self.w, 1, out)
    table.insert(out, self.b)
    return out
end

function Neuron:__tostring()
    return string.format("Neuron(%d)", #self.w)
end

local Layer = class()
function Layer:init(nin, nout)
    self.neurons = arraycomp(function()
        return Neuron(nin)
    end, range(nout))
    
    --print(string.format("Layer with %dx inputs, %dx outputs, %dx params", nin, nout, nout*(1+nin)))
end

function Layer:__call(x)
    local out = {}
    for _,n in ipairs(self.neurons) do
        table.insert(out, n(x))
    end
    return ((#out == 1) and out[1]) or out
end

function Layer:parameters()
    local out = {}
    for _,n in ipairs(self.neurons) do
        local params = n:parameters()
        table.move(params, 1, #params, #out+1, out)
    end
    return out
end

function Layer:__tostring()
    local ns = {}
    for _,n in ipairs(self.neurons) do
        table.insert(ns, tostring(n))
    end
    return string.format("Layer of [%s]", table.concat(ns, ', '))
end

local MLP = class()
function MLP:init(nin, nouts)
    local sz = arrayconcat({nin}, nouts)
    self.layers = arraycomp(function(i)
        return Layer(sz[i], sz[i+1])
    end, range(#nouts))
end

function MLP:__call(x)
    for _,layer in ipairs(self.layers) do
        x = layer(x)
    end
    return x
end

function MLP:parameters()
    local out = {}
    for _,layer in ipairs(self.layers) do
        local params = layer:parameters()
        table.move(params, 1, #params, #out+1, out)
    end
    return out
end

function MLP:__tostring()
    local ls = {}
    for _,l in ipairs(self.layers) do
        table.insert(ls, tostring(l))
    end
    return string.format("MLP of [%s]", table.concat(ls, ', '))
end

AI.Neuron = Neuron
AI.Layer = Layer
AI.MLP = MLP
